"""This module contains global state of flynt application instance."""

verbose = False
quiet = False
aggressive = False
dry_run = False

percent_candidates = 0
percent_transforms = 0

call_candidates = 0
call_transforms = 0

invalid_conversions = 0

concat_candidates = 0
concat_changes = 0
