farewell_message = (
    "Please run your tests before committing. Did flynt get a perfect conversion? give it a star at: "
    "\n~ https://github.com/ikamensh/flynt ~"
    "\nThank you for using flynt. Upgrade more projects and recommend it to your colleagues!\n"
)
